@@ -10,6 +10,8 @@
 type t = string
 exception Invalid of string * string
 
+let fail str pos msg = raise (Invalid(Printf.sprintf "at position %d: %s" pos msg, str))
+
 (* +-----------------------------------------------------------------+
    | Unsafe primitives (use with caution!)                           |
    +-----------------------------------------------------------------+ *)
@@ -26,41 +28,66 @@ let unsafe_sub str ofs len =
    | UTF-8 validation                                                |
    +-----------------------------------------------------------------+ *)
 
-let check str =
-  let len = String.length str in
-  let fail i msg = Some(Printf.sprintf "at offset %d: %s" i msg) in
-  let rec trail i minimum acc = function
-    | 0 ->
-        if acc < minimum then
-          fail i "overlong UTF-8 sequence"
-        else
-          loop i
-    | count ->
-        if i = len then
-          fail i "unterminated UTF-8 sequence"
-        else
-          let n = byte str i in
-          if n land 0xc0 = 0x80 then
-            trail (i + 1) minimum ((acc lsl 6) lor (n land 0x3f)) (count - 1)
-          else
-            fail i "unterminated UTF-8 sequence"
-  and loop i =
+let check s =
+  let fail i msg = Some(Printf.sprintf "at position %d: %s" i msg) in
+  let len = String.length s in
+  let rec main i =
     if i = len then
       None
     else
-      let n = byte str i in
-      if n land 0x80 = 0 then
-        loop (i + 1)
-      else if n land 0xe0 = 0xc0 then
-        trail (i + 1) 0x80 (n land 0x1f) 1
-      else if n land 0xf0 = 0xe0 then
-        trail (i + 1) 0x800 (n land 0x0f) 2
-      else if n land 0xf8 = 0xf0 then
-        trail (i + 1) 0x10000 (n land 0x07) 3
-      else
-        fail i "invalid start of UTF-8 sequence"
+      let ch = String.unsafe_get s i in
+      match ch with
+        | '\x00' .. '\x7f' ->
+            main (i + 1)
+        | '\xc0' .. '\xdf' ->
+            if i + 1 >= len then
+              fail len "premature end of UTF8 sequence"
+            else begin
+              let byte1 = Char.code (String.unsafe_get s (i + 1)) in
+              if byte1 land 0xc0 != 0x80 then
+                fail (i + 1) "malformed UTF8 sequence"
+              else if ((Char.code ch land 0x1f) lsl 6) lor (byte1 land 0x3f) < 0x80 then
+                fail i "overlong UTF8 sequence"
+              else
+                main (i + 2)
+            end
+        | '\xe0' .. '\xef' ->
+            if i + 2 >= len then
+              fail len "premature end of UTF8 sequence"
+            else begin
+              let byte1 = Char.code (String.unsafe_get s (i + 1))
+              and byte2 = Char.code (String.unsafe_get s (i + 2)) in
+              if byte1 land 0xc0 != 0x80 then
+                fail (i + 1) "malformed UTF8 sequence"
+              else if byte2 land 0xc0 != 0x80 then
+                fail (i + 2) "malformed UTF8 sequence"
+              else if ((Char.code ch land 0x0f) lsl 12) lor ((byte1 land 0x3f) lsl 6) lor (byte2 land 0x3f) < 0x800 then
+                fail i "overlong UTF8 sequence"
+              else
+                main (i + 3)
+            end
+        | '\xf0' .. '\xf7' ->
+            if i + 3 >= len then
+              fail len "premature end of UTF8 sequence"
+            else begin
+              let byte1 = Char.code (String.unsafe_get s (i + 1))
+              and byte2 = Char.code (String.unsafe_get s (i + 2))
+              and byte3 = Char.code (String.unsafe_get s (i + 3)) in
+              if byte1 land 0xc0 != 0x80 then
+                fail (i + 1) "malformed UTF8 sequence"
+              else if byte2 land 0xc0 != 0x80 then
+                fail (i + 2) "malformed UTF8 sequence"
+              else if byte3 land 0xc0 != 0x80 then
+                fail (i + 3) "malformed UTF8 sequence"
+              else if ((Char.code ch land 0x07) lsl 18) lor ((byte1 land 0x3f) lsl 12) lor ((byte2 land 0x3f) lsl 6) lor (byte3 land 0x3f) < 0x10000 then
+                fail i "overlong UTF8 sequence"
+              else
+                main (i + 4)
+            end
+        | _ ->
+            fail i "invalid start of UTF8 sequence"
   in
-  loop 0
+  main 0
 
 let invalid str = match check str with
   | None -> raise (Invalid("", str))
@@ -81,6 +108,89 @@ let decode ?(encoding=sys_encoding) txt = Encoding.recode_string ~src:encoding ~
 
 let to_ascii txt = encode ~encoding:"ASCII//TRANSLIT" txt
 
+(* +-----------------------------------------------------------------+
+   | Unsafe UTF-8 naviguation                                        |
+   +-----------------------------------------------------------------+ *)
+
+let unsafe_next str ofs =
+  match String.unsafe_get str ofs with
+    | '\x00' .. '\x7f' ->
+        ofs + 1
+    | '\xc0' .. '\xdf' ->
+        if ofs + 2 > String.length str then
+          fail str ofs "unterminated UTF-8 sequence"
+        else
+          ofs + 2
+    | '\xe0' .. '\xef' ->
+        if ofs + 3 > String.length str then
+          fail str ofs "unterminated UTF-8 sequence"
+        else
+          ofs + 3
+    | '\xf0' .. '\xf7' ->
+        if ofs + 4 > String.length str then
+          fail str ofs "unterminated UTF-8 sequence"
+        else
+          ofs + 4
+    | _ ->
+        fail str ofs "invalid start of UTF-8 sequence"
+
+let unsafe_prev str ofs =
+  match String.unsafe_get str (ofs - 1) with
+    | '\x00' .. '\x7f' ->
+        ofs - 1
+    | '\x80' .. '\xbf' ->
+        if ofs >= 2 then
+          match String.unsafe_get str (ofs - 2) with
+            | '\xc0' .. '\xdf' ->
+                ofs - 2
+            | '\x80' .. '\xbf' ->
+                if ofs >= 3 then
+                  match String.unsafe_get str (ofs - 3) with
+                    | '\xe0' .. '\xef' ->
+                        ofs - 3
+                    | '\x80' .. '\xbf' ->
+                        if ofs >= 4 then
+                          match String.unsafe_get str (ofs - 4) with
+                            | '\xf0' .. '\xf7' ->
+                                ofs - 4
+                            | _ ->
+                                fail str (ofs - 4) "invalid start of UTF-8 sequence"
+                        else
+                          fail str (ofs - 3) "invalid start of UTF-8 string"
+                    | _ ->
+                        fail str (ofs - 3) "invalid middle of UTF-8 sequence"
+                else
+                  fail str (ofs - 2) "invaild start of UTF-8 string"
+            | _ ->
+                fail str (ofs - 2) "invalid middle of UTF-8 sequence"
+        else
+          fail str (ofs - 1) "invalid start of UTF-8 string"
+    | _ ->
+        fail str (ofs - 1) "invalid end of UTF-8 sequence"
+
+let unsafe_extract_next str ofs =
+  let ch = String.unsafe_get str ofs in
+  match ch with
+    | '\x00' .. '\x7f' ->
+        (Char.code ch, ofs + 1)
+    | '\xc0' .. '\xdf' ->
+        if ofs + 2 > String.length str then
+          fail str ofs "unterminated UTF-8 sequence"
+        else
+          (((Char.code ch land 0x1f) lsl 6) lor (byte str (ofs + 1) land 0x3f), ofs + 2)
+    | '\xe0' .. '\xef' ->
+        if ofs + 3 > String.length str then
+          fail str ofs "unterminated UTF-8 sequence"
+        else
+          (((Char.code ch land 0x0f) lsl 12) lor ((byte str (ofs + 1) land 0x3f) lsl 6) lor (byte str (ofs + 2) land 0x3f), ofs + 3)
+    | '\xf0' .. '\xf7' ->
+        if ofs + 4 > String.length str then
+          fail str ofs "unterminated UTF-8 sequence"
+        else
+          (((Char.code ch land 0x07) lsl 18) lor ((byte str (ofs + 1) land 0x3f) lsl 12) lor ((byte str (ofs + 2) land 0x3f) lsl 6) lor (byte str (ofs + 3) land 0x3f), ofs + 4)
+    | _ ->
+        fail str ofs "invalid start of UTF-8 sequence"
+
 (* +-----------------------------------------------------------------+
    | Pointers                                                        |
    +-----------------------------------------------------------------+ *)
@@ -97,60 +207,47 @@ let pointer_r txt = { txt = txt; ofs = String.length txt }
 
 let offset ptr = ptr.ofs
 
-(* Find the offset of the end of a sequence *)
-let rec ofs_next txt ofs len =
-  if ofs < len && byte txt ofs land 0xc0 = 0x80 then
-    ofs_next txt (ofs + 1) len
-  else
-    ofs
-
 let next ptr =
   let len = String.length ptr.txt in
   if ptr.ofs = len then
     None
   else
-    let ofs = ofs_next ptr.txt (ptr.ofs + 1) len in
-    Some(unsafe_sub ptr.txt ptr.ofs (ofs - ptr.ofs), { txt = ptr.txt; ofs = ofs })
-
-(* Find the offset of the start of a sequence *)
-let rec ofs_prev txt ofs =
-  if ofs > 0 && byte txt ofs land 0xc0 = 0x80 then
-    ofs_prev txt (ofs - 1)
-  else
-    ofs
+    let ofs = unsafe_next ptr.txt ptr.ofs in
+    Some(unsafe_sub ptr.txt ptr.ofs (ofs - ptr.ofs), { ptr with ofs = ofs })
 
 let prev ptr =
   if ptr.ofs = 0 then
     None
   else
-    let ofs = ofs_prev ptr.txt (ptr.ofs - 1) in
-    Some(unsafe_sub ptr.txt ofs (ptr.ofs - ofs), { txt = ptr.txt; ofs = ofs })
+    let ofs = unsafe_prev ptr.txt ptr.ofs in
+    Some(unsafe_sub ptr.txt ofs (ptr.ofs - ofs), { ptr with ofs = ofs })
 
-let move idx ptr =
-  let len = String.length ptr.txt in
-  if idx >= 0 then begin
-    let ofs = ref ptr.ofs and idx = ref idx in
-    while !idx > 0 do
-      if !ofs = len then
-        invalid_arg "Text.move"
-      else begin
-        ofs := ofs_next ptr.txt (!ofs + 1) len;
-        decr idx
-      end
-    done;
-    { txt = ptr.txt; ofs = !ofs }
-  end else begin
-    let ofs = ref ptr.ofs and idx = ref idx in
-    while !idx < 0 do
-      if !ofs = 0 then
-        invalid_arg "Text.move"
-      else begin
-        ofs := ofs_prev ptr.txt (!ofs - 1);
-        incr idx
-      end
-    done;
-    { txt = ptr.txt; ofs = !ofs }
-  end
+let rec move_offset_l str ofs len =
+  if len = 0 then
+    ofs
+  else if ofs = String.length str then
+    invalid_arg "Text.move"
+  else
+    move_offset_l str (unsafe_next str ofs) (len - 1)
+
+let rec move_offset_r str ofs len =
+  if len = 0 then
+    ofs
+  else if ofs = 0 then
+    invalid_arg "Text.move"
+  else
+    move_offset_r str (unsafe_prev str ofs) (len + 1)
+
+let move_offset str ofs len =
+  if len = 0 then
+    ofs
+  else if len > 0 then
+    move_offset_l str ofs len
+  else
+    move_offset_r str ofs len
+
+let move len ptr =
+  { ptr with ofs = move_offset ptr.txt ptr.ofs len }
 
 let chunk a b =
   if a.txt != b.txt then
@@ -161,18 +258,22 @@ let chunk a b =
     else
       unsafe_sub a.txt b.ofs (a.ofs - b.ofs)
 
-let pointer_at txt idx =
+let offset_at txt idx =
   if idx < 0 then
-    move idx (pointer_r txt)
+    move_offset_r txt (String.length txt) idx
   else
-    move idx (pointer_l txt)
+    move_offset_l txt 0 idx
 
-let position ptr =
-  let count = ref 0 in
-  for i = 0 to ptr.ofs - 1 do
-    if byte ptr.txt i land 0xc0 <> 0x80 then incr count
-  done;
-  !count
+let pointer_at txt idx =
+  { txt = txt; ofs = offset_at txt idx }
+
+let rec position_rec ptr ofs pos =
+  if ofs >= ptr.ofs then
+    pos
+  else
+    position_rec ptr (unsafe_next ptr.txt ofs) (pos + 1)
+
+let position ptr = position_rec ptr 0 0
 
 let rec ptr_equal_at_aux txt sub ofs = function
   | -1 -> true
@@ -191,42 +292,39 @@ let equal_at ptr sub =
    | High-level functions                                            |
    +-----------------------------------------------------------------+ *)
 
-let length t =
-  let len = ref 0 in
-  for i = 0 to String.length t - 1 do
-    if byte t i land 0xc0 <> 0x80 then incr len
-  done;
-  !len
+let rec length_rec str ofs len =
+  if ofs = String.length str then
+    len
+  else
+    length_rec str (unsafe_next str ofs) (len + 1)
 
-let code txt =
-  let len = String.length txt in
-  if len = 0 then
+let length str = length_rec str 0 0
+
+let code str =
+  if str = "" then
     invalid_arg "Text.code"
   else
-    let n = byte txt 0 in
-    if n land 0x80 = 0 then
-      n
-    else if n land 0xe0 = 0xc0 then
-      if len < 2 then
-        invalid txt
-      else
-        ((n land 0x1f) lsl 6)
-        lor (byte txt 1)
-    else if n land 0xf0 = 0xe0 then
-      if len < 3 then
-        invalid txt
-      else
-        ((n land 0x0f) lsl 12)
-        lor ((byte txt 1 land 0x3f) lsl 6)
-        lor (byte txt 2 land 0x3f)
-    else
-      if len < 4 then
-        invalid txt
-      else
-        ((n land 0x07) lsl 18)
-        lor ((byte txt 1 land 0x3f) lsl 12)
-        lor ((byte txt 2 land 0x3f) lsl 6)
-        lor (byte txt 3 land 0x3f)
+    let ch = String.unsafe_get str 0 in
+    match ch with
+      | '\x00' .. '\x7f' ->
+          Char.code ch
+      | '\xc0' .. '\xdf' ->
+          if 2 > String.length str then
+            fail str 0 "unterminated UTF-8 sequence"
+          else
+            ((Char.code ch land 0x1f) lsl 6) lor (byte str 1 land 0x3f)
+      | '\xe0' .. '\xef' ->
+          if 3 > String.length str then
+            fail str 0 "unterminated UTF-8 sequence"
+          else
+            ((Char.code ch land 0x0f) lsl 12) lor ((byte str 1 land 0x3f) lsl 6) lor (byte str 2 land 0x3f)
+      | '\xf0' .. '\xf7' ->
+          if 4 > String.length str then
+            fail str 0 "unterminated UTF-8 sequence"
+          else
+            ((Char.code ch land 0x07) lsl 18) lor ((byte str 1 land 0x3f) lsl 12) lor ((byte str 2 land 0x3f) lsl 6) lor (byte str 3 land 0x3f)
+      | _ ->
+          fail str 0 "invalid start of UTF-8 sequence"
 
 let char code =
   if code < 0 then
@@ -257,31 +355,39 @@ let char code =
     invalid_arg "Text.char"
 
 let get txt idx =
-  match next (pointer_at txt idx) with
-    | Some(t, _) -> t
-    | None -> invalid_arg "Text.get"
+  let ofs = offset_at txt idx in
+  if ofs = String.length txt then
+    invalid_arg "Text.get"
+  else
+    let ofs' = unsafe_next txt ofs in
+    unsafe_sub txt ofs (ofs' - ofs)
 
 let sub txt idx len =
-  let a = pointer_at txt idx in
-  let b = move len a in
-  if a.ofs > b.ofs then
-    unsafe_sub txt b.ofs (a.ofs - b.ofs)
+  let a = offset_at txt idx in
+  let b = move_offset txt a len in
+  if a > b then
+    unsafe_sub txt b (a - b)
   else
-    unsafe_sub txt a.ofs (b.ofs - a.ofs)
+    unsafe_sub txt a (b - a)
 
 let slice txt a b =
-  let a = pointer_at txt a and b = pointer_at txt b in
-  if a.ofs > b.ofs then
+  let a = offset_at txt a and b = offset_at txt b in
+  if a > b then
     invalid_arg "Text.slice"
   else
-    unsafe_sub txt a.ofs (b.ofs - a.ofs)
+    unsafe_sub txt a (b - a)
 
 let splice txt a b repl =
-  let a = pointer_at txt a and b = pointer_at txt b in
-  if a.ofs > b.ofs then
+  let a = offset_at txt a and b = offset_at txt b in
+  if a > b then
     invalid_arg "Text.slice"
-  else
-    String.concat "" [unsafe_sub txt 0 a.ofs; repl; unsafe_sub txt b.ofs (String.length txt - b.ofs)]
+  else begin
+    let res = String.create (a + String.length repl + String.length txt - b) in
+    String.unsafe_blit txt 0 res 0 a;
+    String.unsafe_blit repl 0 res a (String.length repl);
+    String.unsafe_blit txt b res (a + String.length repl) (String.length txt - b);
+    res
+  end
 
 let repeat n txt =
   let len = String.length txt in
@@ -293,26 +399,30 @@ let repeat n txt =
   done;
   res
 
-let iter f txt =
-  let rec loop ptr = match next ptr with
-    | Some(t, ptr) -> f t; loop ptr
-    | None -> ()
-  in
-  loop (pointer_l txt)
+let rec iter_rec f txt ofs =
+  if ofs <> String.length txt then begin
+    let ofs' = unsafe_next txt ofs in
+    f (unsafe_sub txt ofs (ofs' - ofs));
+    iter_rec f txt ofs'
+  end
 
-let rev_iter f txt =
-  let rec loop ptr = match prev ptr with
-    | Some(t, ptr) -> f t; loop ptr
-    | None -> ()
-  in
-  loop (pointer_r txt)
+let iter f txt = iter_rec f txt 0
+
+let rec rev_iter_rec f txt ofs =
+  if ofs <> 0 then begin
+    let ofs' = unsafe_prev txt ofs in
+    f (unsafe_sub txt ofs' (ofs - ofs'));
+    rev_iter_rec f txt ofs'
+  end
+
+let rev_iter f txt = rev_iter_rec f txt (String.length txt)
 
 let rev txt =
   let len = String.length txt in
   let ofs_src = ref len and ofs_dst = ref 0 in
   let res = String.create len in
   while !ofs_src > 0 do
-    let ofs = ofs_prev txt (!ofs_src - 1) in
+    let ofs = unsafe_prev txt !ofs_src in
     let len = !ofs_src - ofs in
     String.unsafe_blit txt ofs res !ofs_dst len;
     ofs_src := ofs;
@@ -334,8 +444,48 @@ let rev_init n f =
   done;
   Buffer.contents buf
 
-let concat = String.concat
-let rev_concat sep l = String.concat sep (List.rev l)
+
+let concat sep l =
+  match l with
+    | [] ->
+        ""
+    | x :: l ->
+        let sep_len = String.length sep in
+        let len = List.fold_left (fun len str -> len + sep_len + String.length str) (String.length x) l in
+        let res = String.create len in
+        String.unsafe_blit x 0 res 0 (String.length x);
+        ignore
+          (List.fold_left
+             (fun ofs str ->
+                String.unsafe_blit sep 0 res ofs sep_len;
+                let ofs = ofs + sep_len in
+                let len = String.length str in
+                String.unsafe_blit str 0 res ofs len;
+                ofs + len)
+             (String.length x) l);
+        res
+
+let rev_concat sep l =
+  match l with
+    | [] ->
+        ""
+    | x :: l ->
+        let sep_len = String.length sep in
+        let len = List.fold_left (fun len str -> len + sep_len + String.length str) (String.length x) l in
+        let res = String.create len in
+        let ofs = len - String.length x in
+        String.unsafe_blit x 0 res ofs (String.length x);
+        ignore
+          (List.fold_left
+             (fun ofs str ->
+                let ofs = ofs - sep_len in
+                String.unsafe_blit sep 0 res ofs sep_len;
+                let len = String.length str in
+                let ofs = ofs - len in
+                String.unsafe_blit str 0 res ofs len;
+                ofs)
+             ofs l);
+        res
 
 let explode txt =
   let l = ref [] in
@@ -373,19 +523,25 @@ let rev_fold f txt acc =
 let filter f txt = map (fun ch -> if f ch then ch else "") txt
 let rev_filter f txt = rev_map (fun ch -> if f ch then ch else "") txt
 
-let for_all f txt =
-  let rec loop ptr = match next ptr with
-    | None -> true
-    | Some(t, ptr) -> f t && loop ptr
-  in
-  loop (pointer_l txt)
+let rec for_all_rec f txt ofs =
+  if ofs = String.length txt then
+    true
+  else begin
+    let ofs' = unsafe_next txt ofs in
+    f (unsafe_sub txt ofs (ofs' - ofs)) && for_all_rec f txt ofs'
+  end
 
-let exists f txt =
-  let rec loop ptr = match next ptr with
-    | None -> false
-    | Some(t, ptr) -> f t || loop ptr
-  in
-  loop (pointer_l txt)
+let for_all f txt = for_all_rec f txt 0
+
+let rec exists_rec f txt ofs =
+  if ofs = String.length txt then
+    false
+  else begin
+    let ofs' = unsafe_next txt ofs in
+    f (unsafe_sub txt ofs (ofs' - ofs)) || exists_rec f txt ofs'
+  end
+
+let exists f txt = exists_rec f txt 0
 
 let count f txt =
   let c = ref 0 in
@@ -409,7 +565,15 @@ external ml_is_space : int -> bool = "ml_text_is_space"
 external ml_is_upper : int -> bool = "ml_text_is_upper"
 external ml_is_xdigit : int -> bool = "ml_text_is_xdigit"
 
-let for_all_code f txt = for_all (fun ch -> f (code ch)) txt
+let rec for_all_code_rec f txt ofs =
+  if ofs = String.length txt then
+    true
+  else begin
+    let code, ofs' = unsafe_extract_next txt ofs in
+    f code && for_all_code_rec f txt ofs'
+  end
+
+let for_all_code f txt = for_all_code_rec f txt 0
 
 let is_ascii s =
   let rec loop = function
@@ -496,7 +660,7 @@ let split ?(max=max_int) ?(sep=" ") txt =
       if ptr_equal_at txt ofs sep len sep_len then
         String.sub txt start_ofs (ofs - start_ofs) :: loop (ofs + sep_len) (rem - 1)
       else
-        loop_word start_ofs (ofs_next txt (ofs + 1) len) rem
+        loop_word start_ofs (unsafe_next txt ofs) rem
   in
   if sep = "" then
     explode txt
@@ -513,7 +677,7 @@ let rev_split ?(max=max_int) ?(sep=" ") txt =
     if ofs = 0 then
       String.sub txt 0 end_ofs :: acc
     else
-      let ofs = ofs_prev txt (ofs - 1) in
+      let ofs = unsafe_prev txt ofs in
       if ptr_equal_at txt ofs sep len sep_len then
         loop (String.sub txt (ofs + sep_len) (end_ofs - ofs - sep_len) :: acc) ofs (rem - 1)
       else
@@ -533,7 +697,7 @@ let find ?from txt patt =
       if ptr_equal_at txt ofs patt len patt_len then
         Some{ txt = txt; ofs = ofs }
       else
-        loop (ofs_next txt (ofs + 1) len)
+        loop (unsafe_next txt ofs)
   in
   loop (match from with
           | Some ptr -> ptr.ofs
@@ -548,7 +712,7 @@ let rev_find ?from txt patt =
       if ptr_equal_at txt ofs patt len patt_len then
         Some{ txt = txt; ofs = ofs }
       else
-        loop (ofs_prev txt (ofs - 1))
+        loop (unsafe_prev txt ofs)
   in
   loop (match from with
           | Some ptr -> ptr.ofs
@@ -568,7 +732,7 @@ let replace text ~patt ~repl =
         let ofs = ofs + patt_len in
         loop ofs ofs
       end else
-        loop ofs_start (ofs_next text (ofs + 1) len)
+        loop ofs_start (unsafe_next text ofs)
   in
   match patt, text with
     | "", "" ->
@@ -587,7 +751,7 @@ let contains txt = function
         if ofs = len then
           false
         else
-          ptr_equal_at txt ofs sub len sub_len || loop (ofs_next txt (ofs + 1) len)
+          ptr_equal_at txt ofs sub len sub_len || loop (unsafe_next txt ofs)
       in
       loop 0
 
@@ -647,21 +811,32 @@ let rchop = function
   | "" -> ""
   | txt ->
       let len = String.length txt in
-      let ofs = ofs_prev txt (len - 1) in
+      let ofs = unsafe_prev txt len in
       unsafe_sub txt 0 ofs
 
 let lchop = function
   | "" -> ""
   | txt ->
       let len = String.length txt in
-      let ofs = ofs_next txt 1 len in
+      let ofs = unsafe_next txt 0 in
       unsafe_sub txt ofs (len - ofs)
 
 (* +-----------------------------------------------------------------+
    | Upper/lower casing                                              |
    +-----------------------------------------------------------------+ *)
 
-let map_code f txt = map (fun ch -> char (f (code ch))) txt
+let rec map_code_rec f buf txt ofs =
+  if ofs = String.length txt then
+    Buffer.contents buf
+  else begin
+    let code, ofs = unsafe_extract_next txt ofs in
+    Buffer.add_string buf (char (f code));
+    map_code_rec f buf txt ofs
+  end
+
+let map_code f txt =
+  let buf = Buffer.create (String.length txt) in
+  map_code_rec f buf txt 0
 
 external ml_upper : int -> int = "ml_text_upper"
 external ml_lower : int -> int = "ml_text_lower"
@@ -672,9 +847,8 @@ let lower = map_code ml_lower
 let map_first_code f = function
   | "" -> ""
   | txt ->
-      let len = String.length txt in
-      let ptr = ofs_next txt 1 len in
-      char (f (code (unsafe_sub txt 0 ptr))) ^ unsafe_sub txt ptr (len - ptr)
+      let code, ofs = unsafe_extract_next txt 0 in
+      char (f code) ^ unsafe_sub txt ofs (String.length txt - ofs)
 
 let capitalize = map_first_code ml_upper
 let uncapitalize = map_first_code ml_lower
